@@ -18,8 +18,10 @@
 
 /*
  * Estimate space of an XFS filesystem
+ *
+ * XXX: assumes dirv1 format.
  */
-#include <xfs/libxfs.h>
+#include "libxfs.h"
 #include <sys/stat.h>
 #include <ftw.h>
 
@@ -48,7 +50,7 @@ int ffn(const char *, const struct stat64 *, int, struct FTW *);
 #define BLOCKSIZE	4096
 #define INODESIZE	256
 #define PERDIRENTRY	\
-	(sizeof(xfs_dir_leaf_entry_t) + sizeof(xfs_dir_leaf_name_t))
+	(sizeof(xfs_dir2_leaf_entry_t) + sizeof(xfs_dir2_data_entry_t))
 #define LOGSIZE		1000
 
 #define FBLOCKS(n)	((n)/blocksize)
