@@ -16,9 +16,8 @@
  * Inc.,  51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
-#include <xfs/xfs.h>
-#include <xfs/command.h>
-#include <xfs/input.h>
+#include "command.h"
+#include "input.h"
 #include <sys/sendfile.h>
 #include "init.h"
 #include "io.h"
@@ -176,8 +175,8 @@ done:
 void
 sendfile_init(void)
 {
-	sendfile_cmd.name = _("sendfile");
-	sendfile_cmd.altname = _("send");
+	sendfile_cmd.name = "sendfile";
+	sendfile_cmd.altname = "send";
 	sendfile_cmd.cfunc = sendfile_f;
 	sendfile_cmd.argmin = 2;
 	sendfile_cmd.argmax = -1;
