@@ -16,14 +16,9 @@
  * Inc.,  51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
-#include <xfs/libxfs.h>
+#include "libxfs.h"
 #include <ctype.h>
-#ifdef ENABLE_BLKID
 #include <blkid/blkid.h>
-#else
-#include <disk/fstyp.h>
-#include <disk/volume.h>
-#endif
 #include "xfs_mkfs.h"
 
 /*
@@ -35,7 +30,6 @@ struct fs_topology {
 	int	rtswidth;	/* stripe width - rt subvolume */
 	int	lsectorsize;	/* logical sector size &*/
 	int	psectorsize;	/* physical sector size */
-	int	sectoralign;
 };
 
 /*
@@ -109,6 +103,8 @@ char	*iopts[] = {
 	"attr",
 #define	I_PROJID32BIT	6
 	"projid32bit",
+#define I_SPINODES	7
+	"sparse",
 	NULL
 };
 
@@ -195,14 +191,14 @@ char	*mopts[] = {
 /*
  * Use this macro before we have superblock and mount structure
  */
-#define	DTOBT(d)	((xfs_drfsbno_t)((d) >> (blocklog - BBSHIFT)))
+#define	DTOBT(d)	((xfs_rfsblock_t)((d) >> (blocklog - BBSHIFT)))
 
 /*
  * Use this for block reservations needed for mkfs's conditions
  * (basically no fragmentation).
  */
 #define	MKFS_BLOCKRES_INODE	\
-	((uint)(XFS_IALLOC_BLOCKS(mp) + (XFS_IN_MAXLEVELS(mp) - 1)))
+	((uint)(mp->m_ialloc_blks + (mp->m_in_maxlevels - 1)))
 #define	MKFS_BLOCKRES(rb)	\
 	((uint)(MKFS_BLOCKRES_INODE + XFS_DA_NODE_MAXDEPTH + \
 	(XFS_BM_MAXLEVELS(mp, XFS_DATA_FORK) - 1) + (rb)))
@@ -295,7 +291,6 @@ calc_stripe_factors(
 	}
 }
 
-#ifdef ENABLE_BLKID
 /*
  * Check for existing filesystem or partition table on device.
  * Returns:
@@ -462,31 +457,35 @@ static void get_topology(
 	struct fs_topology	*ft,
 	int			force_overwrite)
 {
-	if (!xi->disfile) {
-		char *dfile = xi->volname ? xi->volname : xi->dname;
-		struct stat statbuf;
+	struct stat statbuf;
+	char *dfile = xi->volname ? xi->volname : xi->dname;
 
-		/*
-		 * If our target is a regular file, and xi->disfile isn't
-		 * set (i.e. no "-d file" invocation), use platform_findsizes
-		 * to try to obtain the underlying filesystem's requirements
-		 * for direct IO; we'll set our sector size to that if possible.
-		 */
-		if (!stat(dfile, &statbuf) && S_ISREG(statbuf.st_mode)) {
-			int fd;
-			long long dummy;
+	/*
+	 * If our target is a regular file, use platform_findsizes
+	 * to try to obtain the underlying filesystem's requirements
+	 * for direct IO; we'll set our sector size to that if possible.
+	 */
+	if (xi->disfile ||
+	    (!stat(dfile, &statbuf) && S_ISREG(statbuf.st_mode))) {
+		int fd;
+		int flags = O_RDONLY;
+		long long dummy;
 
-			fd = open(dfile, O_RDONLY);
-			if (fd >= 0) {
-				platform_findsizes(dfile, fd, &dummy,
-						   &ft->lsectorsize);
-				close(fd);
-			}
-		} else {
-			blkid_get_topology(dfile, &ft->dsunit, &ft->dswidth,
-					   &ft->lsectorsize, &ft->psectorsize,
-					   force_overwrite);
-		}
+		/* with xi->disfile we may not have the file yet! */
+		if (xi->disfile)
+			flags |= O_CREAT;
+
+		fd = open(dfile, flags, 0666);
+		if (fd >= 0) {
+			platform_findsizes(dfile, fd, &dummy, &ft->lsectorsize);
+			close(fd);
+			ft->psectorsize = ft->lsectorsize;
+		} else
+			ft->psectorsize = ft->lsectorsize = BBSIZE;
+	} else {
+		blkid_get_topology(dfile, &ft->dsunit, &ft->dswidth,
+				   &ft->lsectorsize, &ft->psectorsize,
+				   force_overwrite);
 	}
 
 	if (xi->rtname && !xi->risfile) {
@@ -496,70 +495,12 @@ static void get_topology(
 				   &lsectorsize, &psectorsize, force_overwrite);
 	}
 }
-#else /* ENABLE_BLKID */
-static int
-check_overwrite(
-	char		*device)
-{
-	char		*type;
-
-	if (device && *device) {
-		if ((type = fstype(device)) != NULL) {
-			fprintf(stderr,
-				_("%s: %s appears to contain an existing "
-				"filesystem (%s).\n"), progname, device, type);
-			return 1;
-		}
-		if ((type = pttype(device)) != NULL) {
-			fprintf(stderr,
-				_("%s: %s appears to contain a partition "
-				"table (%s).\n"), progname, device, type);
-			return 1;
-		}
-	}
-	return 0;
-}
-
-static void get_topology(
-	libxfs_init_t		*xi,
-	struct fs_topology	*ft,
-	int			force_overwrite)
-{
-
-	char *dfile = xi->volname ? xi->volname : xi->dname;
-	int bsz = BBSIZE;
-
-	if (!xi->disfile) {
-		int fd;
-		long long dummy;
-
-		get_subvol_stripe_wrapper(dfile, SVTYPE_DATA,
-				&ft->dsunit, &ft->dswidth, &ft->sectoralign);
-		fd = open(dfile, O_RDONLY);
-		/* If this fails we just fall back to BBSIZE */
-		if (fd >= 0) {
-			platform_findsizes(dfile, fd, &dummy, &bsz);
-			close(fd);
-		}
-	}
-
-	ft->lsectorsize = bsz;
-	ft->psectorsize = bsz;
-
-	if (xi->rtname && !xi->risfile) {
-		int dummy1;
-
-		get_subvol_stripe_wrapper(dfile, SVTYPE_RT, &dummy1,
-					  &ft->rtswidth, &dummy1);
-	}
-}
-#endif /* ENABLE_BLKID */
 
 static void
 fixup_log_stripe_unit(
 	int		lsflag,
 	int		sunit,
-	xfs_drfsbno_t	*logblocks,
+	xfs_rfsblock_t	*logblocks,
 	int		blocklog)
 {
 	__uint64_t	tmp_logblocks;
@@ -589,14 +530,14 @@ fixup_log_stripe_unit(
 	}
 }
 
-static xfs_dfsbno_t
+static xfs_fsblock_t
 fixup_internal_log_stripe(
 	xfs_mount_t	*mp,
 	int		lsflag,
-	xfs_dfsbno_t	logstart,
+	xfs_fsblock_t	logstart,
 	__uint64_t	agsize,
 	int		sunit,
-	xfs_drfsbno_t	*logblocks,
+	xfs_rfsblock_t	*logblocks,
 	int		blocklog,
 	int		*lalign)
 {
@@ -922,7 +863,7 @@ main(
 	int			c;
 	int			daflag;
 	int			dasize;
-	xfs_drfsbno_t		dblocks;
+	xfs_rfsblock_t		dblocks;
 	char			*dfile;
 	int			dirblocklog;
 	int			dirblocksize;
@@ -950,11 +891,11 @@ main(
 	int			ldflag;
 	int			liflag;
 	xfs_agnumber_t		logagno;
-	xfs_drfsbno_t		logblocks;
+	xfs_rfsblock_t		logblocks;
 	char			*logfile;
 	int			loginternal;
 	char			*logsize;
-	xfs_dfsbno_t		logstart;
+	xfs_fsblock_t		logstart;
 	int			logversion;
 	int			lvflag;
 	int			lsflag;
@@ -984,9 +925,9 @@ main(
 	char			*protofile;
 	char			*protostring;
 	int			qflag;
-	xfs_drfsbno_t		rtblocks;
+	xfs_rfsblock_t		rtblocks;
 	xfs_extlen_t		rtextblocks;
-	xfs_drtbno_t		rtextents;
+	xfs_rtblock_t		rtextents;
 	char			*rtextsize;
 	char			*rtfile;
 	char			*rtsize;
@@ -1005,6 +946,7 @@ main(
 	int			crcs_enabled;
 	int			finobt;
 	bool			finobtflag;
+	int			spinodes;
 
 	progname = basename(argv[0]);
 	setlocale(LC_ALL, "");
@@ -1024,7 +966,8 @@ main(
 	logversion = 2;
 	logagno = logblocks = rtblocks = rtextblocks = 0;
 	Nflag = nlflag = nsflag = nvflag = nci = 0;
-	nftype = dirftype = 0;		/* inode type information in the dir */
+	dirftype = 1;		/* inode type information in the dir */
+	nftype = 0;
 	dirblocklog = dirblocksize = 0;
 	dirversion = XFS_DFL_DIR_VERSION;
 	qflag = 0;
@@ -1040,6 +983,7 @@ main(
 	crcs_enabled = 1;
 	finobt = 1;
 	finobtflag = false;
+	spinodes = 0;
 	memset(&fsx, 0, sizeof(fsx));
 
 	memset(&xi, 0, sizeof(xi));
@@ -1361,6 +1305,13 @@ main(
 						illegal(value, "i projid32bit");
 					projid16bit = c ? 0 : 1;
 					break;
+				case I_SPINODES:
+					if (!value || *value == '\0')
+						value = "1";
+					spinodes = atoi(value);
+					if (spinodes < 0 || spinodes > 1)
+						illegal(value, "i spinodes");
+					break;
 				default:
 					unknown('i', value);
 				}
@@ -1527,11 +1478,6 @@ main(
 					if (c < 0 || c > 1)
 						illegal(value, "m crc");
 					crcs_enabled = c;
-					if (nftype && crcs_enabled) {
-						fprintf(stderr,
-_("cannot specify both crc and ftype\n"));
-						usage();
-					}
 					break;
 				case M_FINOBT:
 					if (!value || *value == '\0')
@@ -1605,11 +1551,6 @@ _("cannot specify both crc and ftype\n"));
 					if (nftype)
 						respec('n', nopts, N_FTYPE);
 					dirftype = atoi(value);
-					if (crcs_enabled) {
-						fprintf(stderr,
-_("cannot specify both crc and ftype\n"));
-						usage();
-					}
 					nftype = 1;
 					break;
 				default:
@@ -1764,21 +1705,15 @@ _("Minimum block size for CRC enabled filesystems is %d bytes.\n"),
 			XFS_MIN_CRC_BLOCKSIZE);
 		usage();
 	}
+	if (crcs_enabled && !dirftype) {
+		fprintf(stderr, _("cannot disable ftype with crcs enabled\n"));
+		usage();
+	}
 
 	memset(&ft, 0, sizeof(ft));
 	get_topology(&xi, &ft, force_overwrite);
 
-	if (ft.sectoralign) {
-		/*
-		 * Older Linux software RAID versions want the sector size
-		 * to match the block size to avoid switching I/O sizes.
-		 * For the legacy libdisk case we thus set the sector size to
-		 * match the block size.  For systems using libblkid we assume
-		 * that the kernel is recent enough to not require this and
-		 * ft.sectoralign will never be set.
-		 */
-		sectorsize = blocksize;
-	} else if (!ssflag) {
+	if (!ssflag) {
 		/*
 		 * Unless specified manually on the command line use the
 		 * advertised sector size of the device.  We use the physical
@@ -1806,7 +1741,7 @@ _("switching to logical sector size %d\n"),
 		}
 	}
 
-	if (ft.sectoralign || !ssflag) {
+	if (!ssflag) {
 		sectorlog = libxfs_highbit32(sectorsize);
 		if (loginternal) {
 			lsectorsize = sectorsize;
@@ -1903,6 +1838,12 @@ _("warning: finobt not supported without CRC support, disabled.\n"));
 		finobt = 0;
 	}
 
+	if (spinodes && !crcs_enabled) {
+		fprintf(stderr,
+_("warning: sparse inodes not supported without CRC support, disabled.\n"));
+		spinodes = 0;
+	}
+
 	if (nsflag || nlflag) {
 		if (dirblocksize < blocksize ||
 					dirblocksize > XFS_MAX_BLOCKSIZE) {
@@ -1939,7 +1880,7 @@ _("warning: finobt not supported without CRC support, disabled.\n"));
 				(long long)dbytes, XFS_MIN_BLOCKSIZE);
 			usage();
 		}
-		dblocks = (xfs_drfsbno_t)(dbytes >> blocklog);
+		dblocks = (xfs_rfsblock_t)(dbytes >> blocklog);
 		if (dbytes % blocksize)
 			fprintf(stderr, _("warning: "
 	"data length %lld not a multiple of %d, truncated to %lld\n"),
@@ -1976,7 +1917,7 @@ _("warning: finobt not supported without CRC support, disabled.\n"));
 				(long long)logbytes, XFS_MIN_BLOCKSIZE);
 			usage();
 		}
-		logblocks = (xfs_drfsbno_t)(logbytes >> blocklog);
+		logblocks = (xfs_rfsblock_t)(logbytes >> blocklog);
 		if (logbytes % blocksize)
 			fprintf(stderr,
 	_("warning: log length %lld not a multiple of %d, truncated to %lld\n"),
@@ -1998,7 +1939,7 @@ _("warning: finobt not supported without CRC support, disabled.\n"));
 				(long long)rtbytes, XFS_MIN_BLOCKSIZE);
 			usage();
 		}
-		rtblocks = (xfs_drfsbno_t)(rtbytes >> blocklog);
+		rtblocks = (xfs_rfsblock_t)(rtbytes >> blocklog);
 		if (rtbytes % blocksize)
 			fprintf(stderr,
 	_("warning: rt length %lld not a multiple of %d, truncated to %lld\n"),
@@ -2353,7 +2294,7 @@ an AG size that is one stripe unit smaller, for example %llu.\n"),
 				if ( dblocks % agsize != 0 &&
 				    (dblocks % agsize <
 				    XFS_AG_MIN_BLOCKS(blocklog))) {
-					dblocks = (xfs_drfsbno_t)((agcount - 1) * agsize);
+					dblocks = (xfs_rfsblock_t)((agcount - 1) * agsize);
 					agcount--;
 					ASSERT(agcount != 0);
 				}
@@ -2379,7 +2320,7 @@ an AG size that is one stripe unit smaller, for example %llu.\n"),
 	if ( dblocks % agsize != 0 &&
 	     (dblocks % agsize < XFS_AG_MIN_BLOCKS(blocklog))) {
 		ASSERT(!daflag);
-		dblocks = (xfs_drfsbno_t)((agcount - 1) * agsize);
+		dblocks = (xfs_rfsblock_t)((agcount - 1) * agsize);
 		agcount--;
 		ASSERT(agcount != 0);
 	}
@@ -2581,7 +2522,7 @@ _("size %s specified for log subvolume is too large, maximum is %lld blocks\n"),
 		printf(_(
 		   "meta-data=%-22s isize=%-6d agcount=%lld, agsize=%lld blks\n"
 		   "         =%-22s sectsz=%-5u attr=%u, projid32bit=%u\n"
-		   "         =%-22s crc=%-8u finobt=%u\n"
+		   "         =%-22s crc=%-8u finobt=%u, sparse=%u\n"
 		   "data     =%-22s bsize=%-6u blocks=%llu, imaxpct=%u\n"
 		   "         =%-22s sunit=%-6u swidth=%u blks\n"
 		   "naming   =version %-14u bsize=%-6u ascii-ci=%d ftype=%d\n"
@@ -2590,7 +2531,7 @@ _("size %s specified for log subvolume is too large, maximum is %lld blocks\n"),
 		   "realtime =%-22s extsz=%-6d blocks=%lld, rtextents=%lld\n"),
 			dfile, isize, (long long)agcount, (long long)agsize,
 			"", sectorsize, attrversion, !projid16bit,
-			"", crcs_enabled, finobt,
+			"", crcs_enabled, finobt, spinodes,
 			"", blocksize, (long long)dblocks, imaxpct,
 			"", dsunit, dswidth,
 			dirversion, dirblocksize, nci, dirftype,
@@ -2611,6 +2552,8 @@ _("size %s specified for log subvolume is too large, maximum is %lld blocks\n"),
 	sbp->sb_rextents = rtextents;
 	platform_uuid_generate(&uuid);
 	platform_uuid_copy(&sbp->sb_uuid, &uuid);
+	/* Only in memory; libxfs expects this as if read from disk */
+	platform_uuid_copy(&sbp->sb_meta_uuid, &uuid);
 	sbp->sb_logstart = logstart;
 	sbp->sb_rootino = sbp->sb_rbmino = sbp->sb_rsumino = NULLFSINO;
 	sbp->sb_rextsize = rtextblocks;
@@ -2659,6 +2602,20 @@ _("size %s specified for log subvolume is too large, maximum is %lld blocks\n"),
 		sbp->sb_logsectsize = 0;
 	}
 
+	/*
+	 * Sparse inode chunk support has two main inode alignment requirements.
+	 * First, sparse chunk alignment must match the cluster size. Second,
+	 * full chunk alignment must match the inode chunk size.
+	 *
+	 * Copy the already calculated/scaled inoalignmt to spino_align and
+	 * update the former to the full inode chunk size.
+	 */
+	if (spinodes) {
+		sbp->sb_spino_align = sbp->sb_inoalignmt;
+		sbp->sb_inoalignmt = XFS_INODES_PER_CHUNK * isize >> blocklog;
+		sbp->sb_features_incompat |= XFS_SB_FEAT_INCOMPAT_SPINODES;
+	}
+
 	if (force_overwrite)
 		zero_old_xfs_structures(&xi, sbp);
 
@@ -2678,7 +2635,7 @@ _("size %s specified for log subvolume is too large, maximum is %lld blocks\n"),
 	buf = libxfs_getbuf(mp->m_ddev_targp, XFS_SB_DADDR, XFS_FSS_TO_BB(mp, 1));
 	buf->b_ops = &xfs_sb_buf_ops;
 	memset(XFS_BUF_PTR(buf), 0, sectorsize);
-	libxfs_sb_to_disk((void *)XFS_BUF_PTR(buf), sbp, XFS_SB_ALL_BITS);
+	libxfs_sb_to_disk((void *)XFS_BUF_PTR(buf), sbp);
 	libxfs_writebuf(buf, LIBXFS_EXIT_ON_FAILURE);
 	libxfs_purgebuf(buf);
 
@@ -2730,6 +2687,7 @@ _("size %s specified for log subvolume is too large, maximum is %lld blocks\n"),
 	for (agno = 0; agno < agcount; agno++) {
 		struct xfs_agfl	*agfl;
 		int		bucket;
+		struct xfs_perag *pag = xfs_perag_get(mp, agno);
 
 		/*
 		 * Superblock.
@@ -2739,7 +2697,7 @@ _("size %s specified for log subvolume is too large, maximum is %lld blocks\n"),
 				XFS_FSS_TO_BB(mp, 1));
 		buf->b_ops = &xfs_sb_buf_ops;
 		memset(XFS_BUF_PTR(buf), 0, sectorsize);
-		libxfs_sb_to_disk((void *)XFS_BUF_PTR(buf), sbp, XFS_SB_ALL_BITS);
+		libxfs_sb_to_disk((void *)XFS_BUF_PTR(buf), sbp);
 		libxfs_writebuf(buf, LIBXFS_EXIT_ON_FAILURE);
 
 		/*
@@ -2752,7 +2710,7 @@ _("size %s specified for log subvolume is too large, maximum is %lld blocks\n"),
 		agf = XFS_BUF_TO_AGF(buf);
 		memset(agf, 0, sectorsize);
 		if (agno == agcount - 1)
-			agsize = dblocks - (xfs_drfsbno_t)(agno * agsize);
+			agsize = dblocks - (xfs_rfsblock_t)(agno * agsize);
 		agf->agf_magicnum = cpu_to_be32(XFS_AGF_MAGIC);
 		agf->agf_versionnum = cpu_to_be32(XFS_AGF_VERSION);
 		agf->agf_seqno = cpu_to_be32(agno);
@@ -2761,6 +2719,8 @@ _("size %s specified for log subvolume is too large, maximum is %lld blocks\n"),
 		agf->agf_roots[XFS_BTNUM_CNTi] = cpu_to_be32(XFS_CNT_BLOCK(mp));
 		agf->agf_levels[XFS_BTNUM_BNOi] = cpu_to_be32(1);
 		agf->agf_levels[XFS_BTNUM_CNTi] = cpu_to_be32(1);
+		pag->pagf_levels[XFS_BTNUM_BNOi] = 1;
+		pag->pagf_levels[XFS_BTNUM_CNTi] = 1;
 		agf->agf_flfirst = 0;
 		agf->agf_fllast = cpu_to_be32(XFS_AGFL_SIZE(mp) - 1);
 		agf->agf_flcount = 0;
@@ -2775,8 +2735,8 @@ _("size %s specified for log subvolume is too large, maximum is %lld blocks\n"),
 			agf->agf_longest = cpu_to_be32(agsize -
 				XFS_FSB_TO_AGBNO(mp, logstart) - logblocks);
 		}
-		if (XFS_MIN_FREELIST(agf, mp) > worst_freelist)
-			worst_freelist = XFS_MIN_FREELIST(agf, mp);
+		if (xfs_alloc_min_freelist(mp, pag) > worst_freelist)
+			worst_freelist = xfs_alloc_min_freelist(mp, pag);
 		libxfs_writebuf(buf, LIBXFS_EXIT_ON_FAILURE);
 
 		/*
@@ -2948,8 +2908,10 @@ _("size %s specified for log subvolume is too large, maximum is %lld blocks\n"),
 		/*
 		 * Free INO btree root block
 		 */
-		if (!finobt)
+		if (!finobt) {
+			xfs_perag_put(pag);
 			continue;
+		}
 
 		buf = libxfs_getbuf(mp->m_ddev_targp,
 				XFS_AGB_TO_DADDR(mp, agno, XFS_FIBT_BLOCK(mp)),
@@ -2964,6 +2926,7 @@ _("size %s specified for log subvolume is too large, maximum is %lld blocks\n"),
 			xfs_btree_init_block(mp, buf, XFS_FIBT_MAGIC, 0, 0,
 						agno, 0);
 		libxfs_writebuf(buf, LIBXFS_EXIT_ON_FAILURE);
+		xfs_perag_put(pag);
 	}
 
 	/*
@@ -2998,13 +2961,13 @@ _("size %s specified for log subvolume is too large, maximum is %lld blocks\n"),
 		args.agno = agno;
 		args.alignment = 1;
 		args.pag = xfs_perag_get(mp,agno);
-		c = libxfs_trans_reserve(tp, &tres, worst_freelist, 0);
+		c = -libxfs_trans_reserve(tp, &tres, worst_freelist, 0);
 		if (c)
 			res_failed(c);
 
 		libxfs_alloc_fix_freelist(&args, 0);
 		xfs_perag_put(args.pag);
-		libxfs_trans_commit(tp, 0);
+		libxfs_trans_commit(tp);
 	}
 
 	/*
@@ -3206,7 +3169,7 @@ usage( void )
 			    sectlog=n|sectsize=num\n\
 /* force overwrite */	[-f]\n\
 /* inode size */	[-i log=n|perblock=n|size=num,maxpct=n,attr=0|1|2,\n\
-			    projid32bit=0|1]\n\
+			    projid32bit=0|1,sparse=0|1]\n\
 /* no discard */	[-K]\n\
 /* log subvol */	[-l agnum=n,internal,size=num,logdev=xxx,version=n\n\
 			    sunit=value|su=num,sectlog=n|sectsize=num,\n\
